import uuid

from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import DataImportMode, WorkerRun, WorkerVersion
from arkindex.dataimport.utils import get_default_farm_id
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role
from ponos.models import State, Workflow

RECIPE = '''
tasks:
  first:
    image: alpine
  second:
    parents:
      - first
    image: hello-world
'''

# Mock WorkerRun versions docker_image_id attribute
ARTIFACT_ID = uuid.uuid4()


class TestWorkerRuns(FixtureAPITestCase):
    """
    Test worker runs endpoints and methods
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.dataimport_1 = cls.corpus.imports.create(creator=cls.user, mode=DataImportMode.Workers)
        cls.version_1 = WorkerVersion.objects.get(worker__slug='reco')
        cls.worker_1 = cls.version_1.worker
        cls.repo = cls.worker_1.repository
        cls.run_1 = cls.dataimport_1.worker_runs.create(version=cls.version_1, parents=[])
        cls.configuration_1 = cls.worker_1.configurations.create(name="My config", configuration={"key": "value"})
        worker_version = WorkerVersion.objects.exclude(worker=cls.version_1.worker).first()
        cls.configuration_2 = worker_version.worker.configurations.create(name="Config")
        cls.dataimport_2 = cls.corpus.imports.create(creator=cls.user, mode=DataImportMode.Workers)
        # Add an execution access right on the worker
        cls.worker_1.memberships.create(user=cls.user, level=Role.Contributor.value)

    def test_runs_list_requires_login(self):
        response = self.client.get(reverse('api:worker-run-list', kwargs={'pk': str(self.dataimport_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_runs_list_no_execution_right(self):
        """
        Worker runs attached to a process can be listed even if the user has no execution rigths to workers
        This is due to the fact a user can see a process running on a corpus they have access
        """
        self.worker_1.memberships.all().delete()
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:worker-run-list', kwargs={'pk': str(self.dataimport_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['count'], 1)

    def test_runs_list(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.get(reverse('api:worker-run-list', kwargs={'pk': str(self.dataimport_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 1)
        self.assertCountEqual(data['results'], [{
            'id': str(self.run_1.id),
            'worker_version_id': str(self.version_1.id),
            'dataimport_id': str(self.dataimport_1.id),
            'parents': [],
            'worker': {
                'id': str(self.worker_1.id),
                'name': self.worker_1.name,
                'type': self.worker_1.type,
                'slug': self.worker_1.slug,
            },
            'configuration_id': None,
        }])

    def test_runs_list_filter_dataimport(self):
        run_2 = self.dataimport_2.worker_runs.create(
            version=self.version_1,
            parents=[],
        )

        self.client.force_login(self.user)
        response = self.client.get(reverse('api:worker-run-list', kwargs={'pk': str(self.dataimport_2.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 1)
        self.assertEqual(data['results'][0]['id'], str(run_2.id))

    def test_runs_post_requires_login(self):
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': str(self.dataimport_2.id)}),
            data={'worker_version_id': str(self.version_1.id), 'parents': []}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_runs_post_no_execution_right(self):
        """
        An execution access on the target worker version is required to create a worker run
        """
        self.worker_1.memberships.update(level=Role.Guest.value)
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': str(self.dataimport_2.id)}),
            data={'worker_version_id': str(self.version_1.id), 'parents': []}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version_id': ['You do not have an execution access to this version.']})

    def test_runs_post_invalid_version(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': str(self.dataimport_2.id)}),
            data={'worker_version_id': uuid.uuid4(), 'parents': []}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version_id': ['This version does not exist.']})

    def test_runs_post_invalid_dataimport_id(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': uuid.uuid4()}),
            data={'worker_version_id': str(self.version_1.id), 'parents': []}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {'detail': 'Not found.'})

    def test_runs_post_invalid_dataimport_mode(self):
        img_dataimport = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Images)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': str(img_dataimport.id)}),
            data={'worker_version_id': str(self.version_1.id), 'parents': []}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'dataimport': ['Import mode must be Workers']})

    def test_runs_post_dataimport_already_started(self):
        dataimport = self.corpus.imports.create(
            creator=self.user,
            mode=DataImportMode.Workers,
            workflow=Workflow.objects.create(farm_id=get_default_farm_id(), recipe=RECIPE),
        )

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': str(dataimport.id)}),
            data={'worker_version_id': str(self.version_1.id), 'parents': []}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'__all__': [
            'Cannot create a WorkerRun on a DataImport that has already started'
        ]})

    def test_runs_post_create_worker_run(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(14):
            response = self.client.post(
                reverse('api:worker-run-list', kwargs={'pk': str(self.dataimport_2.id)}),
                data={'worker_version_id': str(self.version_1.id), 'parents': []}, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        pk = data.pop('id')
        self.assertNotEqual(pk, self.run_1.id)
        self.assertDictEqual(data, {
            'worker_version_id': str(self.version_1.id),
            'dataimport_id': str(self.dataimport_2.id),
            'parents': [],
            'worker': {
                'id': str(self.worker_1.id),
                'name': self.worker_1.name,
                'type': self.worker_1.type,
                'slug': self.worker_1.slug,
            },
            'configuration_id': None,
        })

    def test_runs_post_via_repository_right(self):
        self.worker_1.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.worker_1.repository.memberships.create(user=self.user, level=Role.Contributor.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(15):
            response = self.client.post(
                reverse('api:worker-run-list', kwargs={'pk': str(self.dataimport_2.id)}),
                data={'worker_version_id': str(self.version_1.id), 'parents': []}, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_runs_post_empty(self):
        """
        Parents and worker_version_id are required to create a worker run
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': str(self.dataimport_2.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'parents': ['This field is required.'],
            'worker_version_id': ['This field is required.']
        })

    def test_create_run_configuration(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(15):
            response = self.client.post(
                reverse('api:worker-run-list', kwargs={'pk': str(self.dataimport_2.id)}),
                data={
                    'worker_version_id': str(self.version_1.id),
                    'parents': [],
                    'configuration_id': str(self.configuration_1.id)
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        pk = data.pop('id')
        self.assertNotEqual(pk, self.run_1.id)
        self.assertDictEqual(data, {
            'worker_version_id': str(self.version_1.id),
            'dataimport_id': str(self.dataimport_2.id),
            'parents': [],
            'worker': {
                'id': str(self.worker_1.id),
                'name': self.worker_1.name,
                'type': self.worker_1.type,
                'slug': self.worker_1.slug,
            },
            'configuration_id': str(self.configuration_1.id)
        })

    def test_create_run_invalid_configuration(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': str(self.dataimport_2.id)}),
            data={'worker_version_id': str(self.version_1.id), 'parents': [], 'configuration_id': str(self.configuration_2.id)}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'configuration_id': ['The configuration must be part of the same worker.']})

    def test_retrieve_run_requires_login(self):
        response = self.client.get(reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retrieve_run_no_worker_execution_right(self):
        """
        A user can retrieve any worker run if they have an admin access to the process project
        """
        self.worker_1.memberships.update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_retrieve_run_invalid_id(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse('api:worker-run-details', kwargs={'pk': '12341234-1234-1234-1234-123412341234'})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_retrieve_run(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'id': str(self.run_1.id),
            'worker_version_id': str(self.version_1.id),
            'dataimport_id': str(self.dataimport_1.id),
            'parents': [],
            'worker': {
                'id': str(self.worker_1.id),
                'name': self.worker_1.name,
                'type': self.worker_1.type,
                'slug': self.worker_1.slug,
            },
            'configuration_id': None,
        })

    def test_update_run_requires_login(self):
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        run_2 = self.dataimport_1.worker_runs.create(
            version=version_2,
            parents=[],
        )

        response = self.client.patch(
            reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
            data={
                'parents': [str(run_2.id)],
            }, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_run_no_project_admin_right(self):
        """
        A user cannot update a worker run if he has no admin access on its process project
        """
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {'detail': 'You do not have an admin access to the process project.'})

    def test_update_run_invalid_id(self):
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        run_2 = self.dataimport_1.worker_runs.create(
            version=version_2,
            parents=[],
        )

        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:worker-run-details', kwargs={'pk': '12341234-1234-1234-1234-123412341234'}),
            data={
                'parents': [str(run_2.id)],
            }, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update_run_inexistant_parent(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
            data={
                'parents': ['12341234-1234-1234-1234-123412341234'],
            }, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), [
            f"Can't add or update WorkerRun {self.run_1.id} because parents field isn't properly defined. It can be either because"
            " one or several UUIDs don't refer to existing WorkerRuns or either because listed WorkerRuns doesn't belong to the"
            " same DataImport than this WorkerRun."
        ])

    def test_update_run_dataimport_id(self):
        """
        Dataimport field cannot be updated
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'dataimport_id': str(self.dataimport_2.id),
                    'parents': [],
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['dataimport_id'], str(self.dataimport_1.id))
        self.run_1.refresh_from_db()
        self.assertEqual(self.run_1.dataimport.id, self.dataimport_1.id)

    def test_update_run_worker_version_id(self):
        """
        Version field cannot be updated
        """
        self.client.force_login(self.user)
        dla_version = WorkerVersion.objects.get(worker__slug='dla')
        with self.assertNumQueries(8):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'worker_version_id': str(dla_version.id),
                    'parents': [],
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.run_1.id),
            'dataimport_id': str(self.dataimport_1.id),
            'parents': [],
            'worker_version_id': str(self.version_1.id),
            'worker': {
                'id': str(self.worker_1.id),
                'name': self.worker_1.name,
                'type': self.worker_1.type,
                'slug': self.worker_1.slug,
            },
            'configuration_id': None,
        })
        self.run_1.refresh_from_db()
        self.assertNotEqual(self.run_1.version_id, dla_version.id)

    def test_update_run_configuration(self):
        self.client.force_login(self.user)
        self.assertEqual(self.run_1.configuration, None)
        with self.assertNumQueries(9):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': self.run_1.id}),
                data={
                    'configuration_id': str(self.configuration_1.id)
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.run_1.refresh_from_db()
        self.assertEqual(self.run_1.configuration.id, self.configuration_1.id)

    def test_update_run_invalid_configuration(self):
        self.client.force_login(self.user)
        self.assertEqual(self.run_1.configuration, None)
        response = self.client.patch(
            reverse('api:worker-run-details', kwargs={'pk': self.run_1.id}),
            data={'configuration_id': str(self.configuration_2.id)},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'configuration_id': ['The configuration must be part of the same worker.']})

    def test_update_run_dataimport_already_started(self):
        """
        Update dependencies of a worker run is not possible once the dataimport is started
        """
        self.dataimport_1.start()
        self.assertIsNotNone(self.dataimport_1.workflow)
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'parents': [],
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            '__all__': ['Cannot update a WorkerRun on a DataImport that has already started']
        })

    def test_update_run(self):
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        run_2 = self.dataimport_1.worker_runs.create(
            version=version_2,
            parents=[],
        )

        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'parents': [str(run_2.id)],
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.run_1.id),
            'worker_version_id': str(self.version_1.id),
            'dataimport_id': str(self.dataimport_1.id),
            'parents': [str(run_2.id)],
            'worker': {
                'id': str(self.worker_1.id),
                'name': self.worker_1.name,
                'type': self.worker_1.type,
                'slug': self.worker_1.slug,
            },
            'configuration_id': None,
        })

    def test_delete_run_requires_login(self):
        response = self.client.delete(
            reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_run_invalid_id(self):
        self.client.force_login(self.user)
        response = self.client.delete(
            reverse('api:worker-run-details', kwargs={'pk': '12341234-1234-1234-1234-123412341234'})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_delete_run_no_worker_execution_right(self):
        """
        A user can delete a worker run with no right on its worker but an admin access to the process project
        """
        self.worker_1.memberships.update(level=Role.Guest.value)
        self.client.force_login(self.user)
        response = self.client.delete(
            reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(WorkerRun.objects.exists())

    def test_delete_run(self):
        self.client.force_login(self.user)
        response = self.client.delete(
            reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(WorkerRun.objects.exists())

    def test_delete_run_with_children(self):
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        rev_3 = self.repo.revisions.create(
            hash='3',
            message='commit message',
            author='dylan',
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        version_3 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_3,
            configuration={"test": "test3"}
        )
        run_2 = self.dataimport_1.worker_runs.create(
            version=version_2,
            parents=[self.run_1.id],
        )
        run_3 = self.dataimport_1.worker_runs.create(
            version=version_3,
            parents=[self.run_1.id, run_2.id],
        )

        self.assertEqual(len(WorkerRun.objects.all()), 3)
        self.assertTrue(self.run_1.id in run_2.parents)
        self.assertTrue(self.run_1.id in run_3.parents)
        self.client.force_login(self.user)
        response = self.client.delete(
            reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(len(WorkerRun.objects.all()), 2)

        run_2.refresh_from_db()
        run_3.refresh_from_db()
        self.assertEqual(run_2.parents, [])
        self.assertEqual(run_3.parents, [run_2.id])

    def test_delete_run_started_process(self):
        """
        A user shouldn't be able to delete the worker run of a started process
        """
        self.client.force_login(self.user)
        self.dataimport_1.start()
        self.dataimport_1.workflow.tasks.update(state=State.Running)
        response = self.client.delete(
            reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'__all__': ["Cannot update a WorkerRun on a DataImport that has already started"]})

    def test_build_task_recipe_no_parent(self):
        self.version_1.docker_image_id = ARTIFACT_ID
        self.assertDictEqual(self.run_1.build_task_recipe('import', '/data/import/elements.json'), {
            'image': f'my_repo.fake/workers/worker/reco:{str(self.version_1.id)}',
            'command': None,
            'artifact': str(ARTIFACT_ID),
            'parents': ['import'],
            'env': {
                'TASK_ELEMENTS': '/data/import/elements.json',
                'WORKER_VERSION_ID': str(self.version_1.id),
                'ARKINDEX_WORKER_RUN_ID': str(self.run_1.id),
            },
        })

    def test_build_task_recipe_with_chunk(self):
        self.version_1.docker_image_id = ARTIFACT_ID
        self.assertDictEqual(self.run_1.build_task_recipe('import', '/data/import/elements.json', chunk=4), {
            'image': f'my_repo.fake/workers/worker/reco:{str(self.version_1.id)}',
            'command': None,
            'artifact': str(ARTIFACT_ID),
            'parents': ['import'],
            'env': {
                'TASK_ELEMENTS': '/data/import/elements.json',
                'WORKER_VERSION_ID': str(self.version_1.id),
                'ARKINDEX_WORKER_RUN_ID': str(self.run_1.id),
                'ARKINDEX_TASK_CHUNK': '4',
            },
        })

    def test_build_task_recipe_with_parent(self):
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        version_2.docker_image_id = ARTIFACT_ID
        run_2 = self.dataimport_1.worker_runs.create(
            version=version_2,
            parents=[self.run_1.id],
        )

        self.assertDictEqual(run_2.build_task_recipe('import', '/data/import/elements.json'), {
            'image': f'my_repo.fake/workers/worker/reco:{str(version_2.id)}',
            'command': None,
            'artifact': str(ARTIFACT_ID),
            'parents': [f'reco_{str(self.version_1.id)[0:6]}'],
            'env': {
                'TASK_ELEMENTS': '/data/import/elements.json',
                'WORKER_VERSION_ID': str(version_2.id),
                'ARKINDEX_WORKER_RUN_ID': str(run_2.id),
            },
        })

    def test_build_task_recipe_with_parent_and_chunk(self):
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        version_2.docker_image_id = ARTIFACT_ID
        run_2 = self.dataimport_1.worker_runs.create(
            version=version_2,
            parents=[self.run_1.id],
        )

        self.assertDictEqual(run_2.build_task_recipe('import', '/data/import/elements.json', chunk=4), {
            'image': f'my_repo.fake/workers/worker/reco:{str(version_2.id)}',
            'command': None,
            'artifact': str(ARTIFACT_ID),
            'parents': [f'reco_{str(self.version_1.id)[0:6]}_4'],
            'env': {
                'TASK_ELEMENTS': '/data/import/elements.json',
                'WORKER_VERSION_ID': str(version_2.id),
                'ARKINDEX_WORKER_RUN_ID': str(run_2.id),
                'ARKINDEX_TASK_CHUNK': '4',
            },
        })
